//
// IOS interface demo
//

// .........................................................................................................
//
// IOS Ports and Opcodes (from IOS source)
//
// .........................................................................................................
//
// AD1-AD0 = 0x0 (I/O write address = 0xFFFFC): EXECUTE WRITE OPCODE.
//
// Execute the previously stored I/O write opcode with the current data.
// The code of the I/O write operation (Opcode) must be previously stored with a STORE OPCODE operation.
// .........................................................................................................
//
// .........................................................................................................
//
// AD1-AD0 = 0x1 (I/O write address = 0xFFFFD): STORE OPCODE.
//
// Store (write) an "I/O operation code" (Opcode) and reset the exchanged bytes counter.
//
// NOTE 1: An Opcode can be a write or read Opcode, if the I/O operation is read or write.
// NOTE 2: the STORE OPCODE operation must always precede an EXECUTE WRITE OPCODE or EXECUTE READ OPCODE 
//         operation.
// NOTE 3: For multi-byte read opcode (as DATETIME) read sequentially all the data bytes without to send
//         a STORE OPCODE operation before each data byte after the first one.
// .........................................................................................................
//
// .........................................................................................................
//
// Currently defined Opcodes for I/O write operations:
//
//   Opcode     Name            Exchanged bytes
// -------------------------------------------------
// Opcode 0x00  USER LED        1
// Opcode 0x01  SERIAL 1 TX     1
// Opcode 0x02  SETIRQ          1
// Opcode 0x03  GPIOA Write     1
// Opcode 0x04  GPIOB Write     1
// Opcode 0x05  IODIRA Write    1
// Opcode 0x06  IODIRB Write    1
// Opcode 0x07  GPPUA Write     1
// Opcode 0x08  GPPUB Write     1
// Opcode 0x09  SELDISK         1
// Opcode 0x10  SERIAL 2 TX     1
// Opcode 0x0A  SELTRACK        2
// Opcode 0x0B  SELSECT         1  
// Opcode 0x0C  WRITESECT       128/512
// Opcode 0xFF  No operation    1
//
//
// Currently defined Opcodes for I/O read operations:
//
//   Opcode     Name            Exchanged bytes
// -------------------------------------------------
// Opcode 0x80  USER KEY        1
// Opcode 0x81  GPIOA Read      1
// Opcode 0x82  GPIOB Read      1
// Opcode 0x84  DATETIME        7
// Opcode 0x85  ERRDISK         1
// Opcode 0x86  READSECT        128/512
// Opcode 0x87  SDMOUNT         1
// Opcode 0x88  FLUSHBUFF       1 
// Opcode 0xFF  No operation    1
//
// See details in the various Opcodes implementation.
//
// .........................................................................................................
//
// AD1-AD0 = 0x0 (I/O read address = 0xFFFFC): EXECUTE READ OPCODE.
//
// Execute the previously stored I/O read operation with the current data.
// The code of the I/O operation (Opcode) must be previously stored with a STORE OPCODE operation.
//
// NOTE: For multi-byte read opcode (as DATETIME) read sequentially all the data bytes without to send
//       a STORE OPCODE operation before each data byte after the first one.
// .........................................................................................................
//          
//
// AD1-AD0 = 0x1 (I/O read address = 0xFFFFD): SERIAL 1 RX.
//
// Execute a Serial I/O Read operation on serial port 1.
// .........................................................................................................
//
// AD1-AD0 = 0x2 (I/O read address = 0xFFFFE): SYSFLAGS.
//
// Read various system flags from the SYSFLAGS register.
// .........................................................................................................
//
// AD1-AD0 = 0x3 (I/O read address = 0xFFFFF): SERIAL 2 RX.
//
// Execute a Serial I/O Read operation on serial port 2.
// .........................................................................................................

#define IOBASE      0xFFFFC                     // Address base for the I/O ports
#define EXCWR_PORT  IOBASE+0                    // Address of the EXECUTE WRITE OPCODE write port
#define EXCRD_PORT  IOBASE+0                    // Address of the EXECUTE READ OPCODE read port
#define STOPC_PORT  IOBASE+1                    // Address of the STORE OPCODE write port
#define SER1RX_PORT IOBASE+1                    // Address of the SERIAL 1 RX read port 
#define SYSFLG_PORT IOBASE+2                    // Address of the SYSFLAGS read port
#define SER2RX_PORT IOBASE+3                    // Address of the SERIAL 2 RX read port
#define USRLED_OPC  0x00                        // USER LED opcode
#define SER1TX_OPC  0x01                        // SERIAL 1 TX opcode
#define SELDISK_OPC 0x09                        // SELDISK opcode
#define SER2TX_OPC  0x10                        // SERIAL 2 TX opcode
#define USERKEY_OPC 0x80                        // USER KEY opcode
#define DATETIM_OPC 0x84                        // DATETIME opcode

#include <stdio.h>

void iosStoreOpc(unsigned char opcode);
void iosExecWriteOpc(unsigned char iodata);
unsigned char iosExecReadOpc(void);
unsigned char iosSysflags(void);
unsigned char iosSer1Rx(void);
unsigned char iosSer2Rx(void);
unsigned char userKey(void);
void userLed(unsigned char c);
void dateTime(unsigned char * vect);
unsigned char delay(unsigned long loop);

int main(void)
{
    unsigned char i;
    unsigned char dateTimeVec[7];

    //
    // Disable buffering on stdout to avoid printf not flushing unless a newline (LF) is sent.
    // For more info: 
    // https://stackoverflow.com/questions/1716296/why-does-printf-not-flush-after-the-call-unless-a-newline-is-in-the-format-strin
    //
    setvbuf(stdout, NULL, _IONBF, 0);

    printf("Press User key to continue... ");
    while (!userKey());
    printf("\n\rUser key pressed\n\r");
    printf("\nNow printing some info:\n\r");

    // Check RTC
    i = iosSysflags();
    if (i & 0b00000010) 
    {
        printf("RTC found\n\r");
        dateTime(dateTimeVec);
        printf("Time: %u/%u/%u\n\r", dateTimeVec[2], dateTimeVec[1], dateTimeVec[0]);
        printf("Date: %u/%u/%u\n\r", dateTimeVec[3], dateTimeVec[4], dateTimeVec[5] + 2000);
        if (dateTimeVec[6] >= 128) dateTimeVec[6] = dateTimeVec[6] -256;    // Two complement correction
        printf("Temperature: %uC\n\r", dateTimeVec[6]);
    }
    else printf("RTC not found\n\r");

    // Check Full/Lite HW config
    if (i & 0b00001000) printf("Lite HW configuration found\n\r");
    else printf("Full HW configuration found\n\r");

    // Blink User led
    printf("\nNow blinking User led...");
    i = 1;
    while (1)
    {
        userLed(i);                         // Toggle User led
        delay(10000);
        i=!i;
    }
}

// ---------------------------------------------------------------------------------------------------------------------
//
// IOS low level interface functions
//
// ---------------------------------------------------------------------------------------------------------------------

// IOS Store Opcode
void iosStoreOpc(unsigned char opcode)
{
    // https://gcc.gnu.org/onlinedocs/gcc-4.8.5/gcc.pdf 
    // par. 6.41  Assembler Instructions with C Expression Operands
    asm("move.b %0,(0xFFFFD).l\n\t"        // move.b  <opcode>,(STOPC_PORT).l           ; Write <opcode> to STOPC_PORT
        : : "r" (opcode)); 
}

// IOS Exec Write Opcode
void iosExecWriteOpc(unsigned char iodata)
{
    // https://gcc.gnu.org/onlinedocs/gcc-4.8.5/gcc.pdf 
    // par. 6.41  Assembler Instructions with C Expression Operands
    asm("move.b %0,(0xFFFFC).l\n\t"        // move.b  <iodata>,(EXCWR_PORT).l           ; Write <iodata> to EXCWR_PORT
        : : "r" (iodata)); 
}

// IOS Exec Read Opcode
unsigned char iosExecReadOpc(void)
{
    volatile unsigned char iodata;

    // https://gcc.gnu.org/onlinedocs/gcc-4.8.5/gcc.pdf 
    // par. 6.41  Assembler Instructions with C Expression Operands
    asm("move.b (0xFFFFC).l,%0\n\t"        // move.b  (EXCRD_PORT).l, <iodata>          ; Read <iodata> from EXCRD_PORT
        : "=r" (iodata));
    return iodata;
}

// IOS read SYSFLAGS
unsigned char iosSysflags(void)
{
    volatile unsigned char iodata;

    // https://gcc.gnu.org/onlinedocs/gcc-4.8.5/gcc.pdf 
    // par. 6.41  Assembler Instructions with C Expression Operands
    asm("move.b (0xFFFFE).l,%0\n\t"        // move.b  (SYSFLG_PORT).l, <iodata>         ; Read <iodata> from SYSFLG_PORT
        : "=r" (iodata));
    return iodata;
}

// IOS SERIAL 1 RX
unsigned char iosSer1Rx(void)
{
    volatile unsigned char iodata;

    // https://gcc.gnu.org/onlinedocs/gcc-4.8.5/gcc.pdf 
    // par. 6.41  Assembler Instructions with C Expression Operands
    asm("move.b (0xFFFFD).l,%0\n\t"        // move.b  (SER1RX_PORT).l, <iodata>         ; Read <iodata> from SER1RX_PORT
        : "=r" (iodata));
    return iodata;
}

// IOS SERIAL 2 RX
unsigned char iosSer2Rx(void)
{
    volatile unsigned char iodata;

    // https://gcc.gnu.org/onlinedocs/gcc-4.8.5/gcc.pdf 
    // par. 6.41  Assembler Instructions with C Expression Operands
    asm("move.b (0xFFFFF).l,%0\n\t"        // move.b  (SER2RX_PORT).l, <iodata>         ; Read <iodata> from SER2RX_PORT
        : "=r" (iodata));
    return iodata;
}

// ---------------------------------------------------------------------------------------------------------------------
//
// IOS Opcode functions (only the ones used here)
//
// ---------------------------------------------------------------------------------------------------------------------

// User key
unsigned char userKey(void)
{
    iosStoreOpc(USERKEY_OPC);
    return iosExecReadOpc();
}

// User led
void userLed(unsigned char status) 
{
    iosStoreOpc(USRLED_OPC);
    iosExecWriteOpc(status);
}

// RTC
void dateTime(unsigned char * vect)
{
    unsigned char i;

    iosStoreOpc(DATETIM_OPC);
    for (i=0; i<7; i++)
    {
        vect[i] = iosExecReadOpc();
    }
}

// ---------------------------------------------------------------------------------------------------------------------
//
// General functions
//
// ---------------------------------------------------------------------------------------------------------------------

// Delay while checking User key
unsigned char delay(unsigned long loop)
{
    unsigned long i;
    unsigned char j = 0;
    for (i=0; i<loop; i++)
    {
        if (userKey() > 0) j = 1;
    }
    return j;
}
