//
// SPP Adapter board Status Lines test: read continuously the values of all the Status Lines (from the printer)
//
// --> REQUIRED: IOS S310121-R231021 (or following revisions until stated otherwise)

// NOTE: the SPP Adapter board must be connected

#define IOBASE      0xFFFFC                     // Address base for the I/O ports
#define EXCWR_PORT  IOBASE+0                    // Address of the EXECUTE WRITE OPCODE write port
#define EXCRD_PORT  IOBASE+0                    // Address of the EXECUTE READ OPCODE read port
#define STOPC_PORT  IOBASE+1                    // Address of the STORE OPCODE write port
#define SER1RX_PORT IOBASE+1                    // Address of the SERIAL 1 RX read port 
#define SYSFLG_PORT IOBASE+2                    // Address of the SYSFLAGS read port
#define SER2RX_PORT IOBASE+3                    // Address of the SERIAL 2 RX read port
#define USRLED_OPC  0x00                        // USER LED opcode
#define SER1TX_OPC  0x01                        // SERIAL 1 TX opcode
#define SELDISK_OPC 0x09                        // SELDISK opcode
#define SER2TX_OPC  0x10                        // SERIAL 2 TX opcode
#define SETSPP_OPC  0x11                        // SETSPP opcode
#define WRSPP_OPC   0x12                        // WRSPP opcode
#define USERKEY_OPC 0x80                        // USER KEY opcode
#define GETSPP_OPC  0x83                        // GETSPP opcode
#define DATETIM_OPC 0x84                        // DATETIME opcode

#include <stdio.h>

void iosStoreOpc(unsigned char opcode);
void iosExecWriteOpc(unsigned char iodata);
unsigned char iosExecReadOpc(void);
unsigned char iosSysflags(void);
unsigned char iosSer1Rx(void);
unsigned char iosSer2Rx(void);
unsigned char getSPP(void);
void setSPP(unsigned char ctrl);
void wrSPP(unsigned char data);

int main(void)
{
    unsigned char i;

    //
    // Disable buffering on stdout to avoid printf not flushing unless a newline (LF) is sent.
    // For more info: 
    // https://stackoverflow.com/questions/1716296/why-does-printf-not-flush-after-the-call-unless-a-newline-is-in-the-format-strin
    //
    setvbuf(stdout, NULL, _IONBF, 0);

    setSPP(0);                                  // Set SPP mode
    if (getSPP() == 0)
    // Failed to set SPP
    {
        printf("GPE not installed. No possible set SPP mode.\n\r");
        return 0;
    }
    printf("GPIO set into SPP mode.\n\r");
    printf("SPP Status lines:\n\r");
    do
    // Read and print continuously the values of the SPP Status Lines
    {
        i = getSPP();                           // Read SPP Status lines
        printf("nACK = %u BUSY = %u PE = %u SELECT = %u nERROR = %u\n\r", (i & 0b00001000)>>3, (i & 0b00010000)>>4, (i & 0b00100000)>>5, (i & 0b01000000)>>6, (i & 0b10000000)>>7);
    }
    while (1);
}


// ---------------------------------------------------------------------------------------------------------------------
//
// IOS low level interface functions
//
// ---------------------------------------------------------------------------------------------------------------------

// IOS Store Opcode
void iosStoreOpc(unsigned char opcode)
{
    // https://gcc.gnu.org/onlinedocs/gcc-4.8.5/gcc.pdf 
    // par. 6.41  Assembler Instructions with C Expression Operands
    asm("move.b %0,(0xFFFFD).l\n\t"        // move.b  <opcode>,(STOPC_PORT).l           ; Write <opcode> to STOPC_PORT
        : : "r" (opcode)); 
}

// IOS Exec Write Opcode
void iosExecWriteOpc(unsigned char iodata)
{
    // https://gcc.gnu.org/onlinedocs/gcc-4.8.5/gcc.pdf 
    // par. 6.41  Assembler Instructions with C Expression Operands
    asm("move.b %0,(0xFFFFC).l\n\t"        // move.b  <iodata>,(EXCWR_PORT).l           ; Write <iodata> to EXCWR_PORT
        : : "r" (iodata)); 
}

// IOS Exec Read Opcode
unsigned char iosExecReadOpc(void)
{
    volatile unsigned char iodata;

    // https://gcc.gnu.org/onlinedocs/gcc-4.8.5/gcc.pdf 
    // par. 6.41  Assembler Instructions with C Expression Operands
    asm("move.b (0xFFFFC).l,%0\n\t"        // move.b  (EXCRD_PORT).l, <iodata>          ; Read <iodata> from EXCRD_PORT
        : "=r" (iodata));
    return iodata;
}

// IOS read SYSFLAGS
unsigned char iosSysflags(void)
{
    volatile unsigned char iodata;

    // https://gcc.gnu.org/onlinedocs/gcc-4.8.5/gcc.pdf 
    // par. 6.41  Assembler Instructions with C Expression Operands
    asm("move.b (0xFFFFE).l,%0\n\t"        // move.b  (SYSFLG_PORT).l, <iodata>         ; Read <iodata> from SYSFLG_PORT
        : "=r" (iodata));
    return iodata;
}

// IOS SERIAL 1 RX
unsigned char iosSer1Rx(void)
{
    volatile unsigned char iodata;

    // https://gcc.gnu.org/onlinedocs/gcc-4.8.5/gcc.pdf 
    // par. 6.41  Assembler Instructions with C Expression Operands
    asm("move.b (0xFFFFD).l,%0\n\t"        // move.b  (SER1RX_PORT).l, <iodata>         ; Read <iodata> from SER1RX_PORT
        : "=r" (iodata));
    return iodata;
}

// IOS SERIAL 2 RX
unsigned char iosSer2Rx(void)
{
    volatile unsigned char iodata;

    // https://gcc.gnu.org/onlinedocs/gcc-4.8.5/gcc.pdf 
    // par. 6.41  Assembler Instructions with C Expression Operands
    asm("move.b (0xFFFFF).l,%0\n\t"        // move.b  (SER2RX_PORT).l, <iodata>         ; Read <iodata> from SER2RX_PORT
        : "=r" (iodata));
    return iodata;
}


// ---------------------------------------------------------------------------------------------------------------------
//
// IOS Opcodes functions (only the ones used here)
//
// ---------------------------------------------------------------------------------------------------------------------

// Read the Status Lines of the SPP Port and the SPP emulation status
unsigned char getSPP(void)
{
    iosStoreOpc(GETSPP_OPC);
    return iosExecReadOpc();
}

// Set the GPIO port into SPP mode
void setSPP(unsigned char ctrl) 
{
    iosStoreOpc(SETSPP_OPC);
    iosExecWriteOpc(ctrl);
}

// Send a byte to the printer attached to the SPP port
void wrSPP(unsigned char data) 
{
    iosStoreOpc(WRSPP_OPC);
    iosExecWriteOpc(data);
}
