//
// Demo using the date/time functions of the C standard library
//

#include <stdio.h>
#include <time.h>

// IOS definitions
#define IOBASE      0xFFFFC                     // Address base for the I/O ports
#define EXCWR_PORT  IOBASE+0                    // Address of the EXECUTE WRITE OPCODE write port
#define EXCRD_PORT  IOBASE+0                    // Address of the EXECUTE READ OPCODE read port
#define STOPC_PORT  IOBASE+1                    // Address of the STORE OPCODE write port
#define SER1RX_PORT IOBASE+1                    // Address of the SERIAL 1 RX read port 
#define SYSFLG_PORT IOBASE+2                    // Address of the SYSFLAGS read port
#define SER2RX_PORT IOBASE+3                    // Address of the SERIAL 2 RX read port
#define DATETIM_OPC 0x84                        // DATETIME opcode

int gettimeofday(struct timeval *tv, void *tz); // NOTE: "struct timeval" is defined in include/sys/_timeval.h
unsigned char iosExecReadOpc(void);
void iosStoreOpc(unsigned char opcode);

int main () 
{
   time_t seconds;

   //
   // Disable buffering on stdout to avoid printf not flushing unless a newline (LF) is sent.
   // For more info: 
   // https://stackoverflow.com/questions/1716296/why-does-printf-not-flush-after-the-call-unless-a-newline-is-in-the-format-strin
   //
   setvbuf(stdout, NULL, _IONBF, 0);

   // Read current date/time as Unix time
   seconds = time(NULL);
   printf("Seconds since January 1, 1970 = %ld\n\r", seconds);

   // Convert it as readble format
   char s[30];
   struct tm tim;
   tim = *(localtime(&seconds));
   strftime(s,30,"%b %d, %Y; %H:%M:%S",&tim);
   printf("Date/Time: %s\n\r", s);
 }

// IOS Store Opcode
void iosStoreOpc(unsigned char opcode)
// IOS low level function - IOS Store Opcode
{
    // https://gcc.gnu.org/onlinedocs/gcc-4.8.5/gcc.pdf 
    // par. 6.41  Assembler Instructions with C Expression Operands
    asm("move.b %0,(0xFFFFD).l\n\t"        // move.b  <opcode>,(STOPC_PORT).l           ; Write <opcode> to STOPC_PORT
        : : "r" (opcode)); 
}

// IOS Exec Read Opcode
unsigned char iosExecReadOpc(void)
// IOS low level function - IOS Exec Read Opcode
{
    volatile unsigned char iodata;

    // https://gcc.gnu.org/onlinedocs/gcc-4.8.5/gcc.pdf 
    // par. 6.41  Assembler Instructions with C Expression Operands
    asm("move.b (0xFFFFC).l,%0\n\t"        // move.b  (EXCRD_PORT).l, <iodata>          ; Read <iodata> from EXCRD_PORT
        : "=r" (iodata));
    return iodata;
}

int gettimeofday(struct timeval *tv, void *tz)
// ****************************************************************************************
// 68k-MBC date/time implementation (gettimeofday).
//
// This function is needed to use the date/time related functions of the C standard library.
// It reads the date/time from the RTC and convert it into Unix time.
//
// NOTE 1: if the RTC is not present the date/time will be set at: Nov 30, 1999; 00:00:00
//
// NOTE 2: this function use the IOS low level functions iosStoreOpc and iosExecReadOpc
// ****************************************************************************************
{
    struct tm info;

    // Read the RTC
    iosStoreOpc(DATETIM_OPC);
    info.tm_sec = iosExecReadOpc();
    info.tm_min = iosExecReadOpc();
    info.tm_hour = iosExecReadOpc();
    info.tm_mday = iosExecReadOpc();
    info.tm_mon = iosExecReadOpc() - 1;             // Month, where 0 = jan
    info.tm_year = iosExecReadOpc() +100;           // Year - 1900
    info.tm_isdst = -1;                             // Is DST on? 1 = yes, 0 = no, -1 = unknown

    // Convert into the Unix time
    tv->tv_sec = mktime(&info);                     // Seconds from Unix epoch
    tv->tv_usec = 0;                                // No microseconds
    return 0;                                       // Return non-zero for error
} 
